'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var compositionApi = require('@vue/composition-api');

function unwrap(o) {
    return compositionApi.isRef(o) ? o.value : o;
}
function wrap(o) {
    return compositionApi.isRef(o) ? o : compositionApi.ref(o);
}
const isArray = Array.isArray;
const isFunction = (val) => typeof val === "function";
const isString = (val) => typeof val === "string";
const isSymbol = (val) => typeof val === "symbol";
const isBoolean = (val) => typeof val === "boolean";
const isDate = (val) => isObject(val) && isFunction(val.getTime);
const isNumber = (val) => typeof val === "number";
const isObject = (val) => val !== null && typeof val === "object";
const isElement = (val) => isObject(val) && !!val.tagName;
function isPromise(val) {
    return isObject(val) && isFunction(val.then) && isFunction(val.catch);
}
const NO_OP = () => { };
const FALSE_OP = () => false;
const PASSIVE_EV = { passive: true };
function promisedTimeout(timeout) {
    return new Promise(res => {
        setTimeout(res, timeout);
    });
}
function minMax(val, min, max) {
    if (val < min)
        return min;
    if (val > max)
        return max;
    return val;
}
const isClient = typeof window != "undefined";
function deepClone(result, ...sources) {
    for (let i = 0; i < sources.length; i++) {
        const source = sources[i];
        if (source === undefined || !isObject(source))
            continue;
        const keys = Object.keys(source);
        for (let j = 0; j < keys.length; j++) {
            const k = keys[j];
            const v = unwrap(source[k]);
            const sourceType = typeof v;
            const type = typeof result[k];
            if (result[k] === undefined || sourceType === type) {
                result[k] = isObject(v)
                    ? deepClone(result[k] || {}, v)
                    : source[k]; // source[k] is assigned because if is ref we want to override to this ref
            }
        }
    }
    return result;
}
// compact version: https://stackoverflow.com/a/33146982/1209882
/**
 * returns a random string
 * @param len length of the string max: 36
 */
// export function randomString(len: number) {
//   return (+new Date).toString(36).slice(-len);
// }

function usePagination(options) {
    const _currentPage = wrap(options.currentPage);
    const _pageSize = wrap(options.pageSize);
    const _offset = compositionApi.ref(0);
    const total = wrap(options.total);
    const offset = compositionApi.computed({
        get() {
            return _offset.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _offset.value = Math.min(v, total.value);
        }
    });
    const currentPage = compositionApi.computed({
        get() {
            return _currentPage.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _currentPage.value = minMax(v, 1, lastPage.value);
            // set the offset
            offset.value = (_currentPage.value - 1) * pageSize.value;
        }
    });
    const pageSize = compositionApi.computed({
        get() {
            return _pageSize.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _pageSize.value = v;
        }
    });
    const lastPage = compositionApi.computed(() => Math.ceil(total.value / pageSize.value));
    // make sure the current page is the correct value
    currentPage.value = _currentPage.value;
    const prev = () => --currentPage.value;
    const next = () => ++currentPage.value;
    const first = () => (currentPage.value = 1);
    const last = () => (currentPage.value = lastPage.value);
    compositionApi.watch([total, pageSize], () => {
        if (currentPage.value > lastPage.value) {
            currentPage.value = lastPage.value;
        }
    }, { lazy: true } // no need to run on first render
    );
    return {
        // Mutable state
        pageSize,
        total,
        currentPage,
        offset,
        // Computed
        lastPage,
        // Functions
        next,
        prev,
        first,
        last
    };
}

function useArrayPagination(array, options) {
    const arrayRef = wrap(array);
    const pagination = usePagination({
        ...{
            currentPage: 1,
            pageSize: 10
        },
        ...options,
        total: compositionApi.computed(() => arrayRef.value.length)
    });
    const result = compositionApi.computed(() => {
        const array = arrayRef.value;
        if (!Array.isArray(array))
            return [];
        return array.slice(pagination.offset.value, pagination.offset.value + pagination.pageSize.value);
    });
    return {
        ...pagination,
        result
    };
}

function usePromise(fn, throwException = false) {
    if (!fn) {
        throw new Error(`[usePromise] argument can't be '${fn}'`);
    }
    if (typeof fn !== "function") {
        throw new Error(`[usePromise] expects function, but received ${typeof fn}`);
    }
    const loading = compositionApi.ref(false);
    const error = compositionApi.ref(null);
    const result = compositionApi.ref(null);
    const promise = compositionApi.ref();
    const exec = async (...args) => {
        loading.value = true;
        error.value = null;
        result.value = null;
        let throwExp = args &&
            fn.length !== args.length &&
            args.length > 0 &&
            isBoolean(args[args.length - 1])
            ? args[args.length - 1]
            : throwException;
        const currentPromise = (promise.value = fn(...args));
        try {
            const r = await currentPromise;
            if (promise.value === currentPromise) {
                result.value = r;
            }
            return r;
        }
        catch (er) {
            if (promise.value === currentPromise) {
                error.value = er;
                result.value = null;
            }
            return throwExp ? currentPromise : undefined;
        }
        finally {
            if (promise.value === currentPromise) {
                loading.value = false;
            }
        }
    };
    return {
        exec,
        result,
        promise,
        loading,
        error
    };
}

function useCancellablePromise(fn, throwException = false) {
    const cancelled = compositionApi.ref(false);
    let _cancel = undefined;
    const cancel = (result) => _cancel(result); // TODO add warning if cancel is undefined
    const promise = (p) => new Promise((res, rej) => {
        _cancel = result => {
            cancelled.value = true;
            rej(result);
        };
        p.then(res).catch(rej);
    });
    const use = usePromise((...args) => promise(fn(...args)), throwException);
    return {
        ...use,
        cancel,
        cancelled
    };
}

const MAX_RETRIES = 9000;
/* istanbul ignore next */
const ExecutionId = Symbol( ``);
/* istanbul ignore next */
const CancellationToken = Symbol( ``);
const defaultStrategy = async (options, context, factory, args) => {
    const retryId = context[ExecutionId].value;
    let i = -1;
    const maxRetries = options.maxRetries || MAX_RETRIES + 1;
    const delay = options.retryDelay || noDelay;
    context.retryErrors.value = [];
    context.isRetrying.value = false;
    context.nextRetry.value = undefined;
    let nextRetry = undefined;
    do {
        let success = false;
        let result = null;
        try {
            ++i;
            if (args) {
                result = factory(...args);
            }
            else {
                result = factory();
            }
            if (isPromise(result)) {
                result = await result;
            }
            // is cancelled?
            if (context[CancellationToken].value) {
                return null;
            }
            success = true;
        }
        catch (error) {
            result = null;
            context.retryErrors.value.push(error);
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
        if (success) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return result;
        }
        if (i >= maxRetries) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return Promise.reject(new Error(`[useRetry] max retries reached ${maxRetries}`));
        }
        context.isRetrying.value = true;
        const now = Date.now();
        const pDelayBy = delay(i); // wrapped
        const delayBy = isPromise(pDelayBy) ? await pDelayBy : pDelayBy; // unwrap promise
        if (!isPromise(pDelayBy) || !!delayBy) {
            if (isNumber(delayBy)) {
                nextRetry = delayBy;
            }
            else if (isDate(delayBy)) {
                nextRetry = delayBy.getTime();
            }
            else {
                throw new Error(`[useRetry] invalid value received from options.retryDelay '${typeof delayBy}'`);
            }
            // if the retry is in the past, means we need to await that amount
            if (nextRetry < now) {
                context.nextRetry.value = now + nextRetry;
            }
            else {
                context.nextRetry.value = nextRetry;
                nextRetry = nextRetry - now;
            }
            if (nextRetry > 0) {
                await promisedTimeout(nextRetry);
            }
        }
        // is cancelled?
        if (context[CancellationToken].value) {
            return null;
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
    } while (i < MAX_RETRIES);
    return null;
};
function useRetry(options, factory) {
    const opt = !options || isFunction(options) ? {} : options;
    const fn = isFunction(options) ? options : factory;
    if (!isFunction(options) && !isObject(options)) {
        throw new Error("[useRetry] options needs to be 'object'");
    }
    if (!!fn && !isFunction(fn)) {
        throw new Error("[useRetry] factory needs to be 'function'");
    }
    const isRetrying = compositionApi.ref(false);
    const nextRetry = compositionApi.ref();
    const retryErrors = compositionApi.ref([]);
    const cancellationToken = { value: false };
    const retryId = { value: 0 };
    const retryCount = compositionApi.computed(() => retryErrors.value.length);
    const context = {
        isRetrying,
        retryCount,
        nextRetry,
        retryErrors,
        [ExecutionId]: retryId,
        [CancellationToken]: cancellationToken
    };
    const exec = fn
        ? (...args) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, fn, args);
        }
        : (f) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, f, undefined);
        };
    const cancel = () => {
        context.isRetrying.value = false;
        context.retryErrors.value.push(new Error("[useRetry] cancelled"));
        context.nextRetry.value = undefined;
        cancellationToken.value = true;
    };
    return {
        ...context,
        cancel,
        exec
    };
}
const exponentialDelay = retryNumber => {
    const delay = Math.pow(2, retryNumber) * 100;
    const randomSum = delay * 0.2 * Math.random(); // 0-20% of the delay
    return delay + randomSum;
};
const noDelay = () => 0;

function useDebounce(handler, wait, options) {
    return debounce(handler, wait, options);
}
function debounce(func, waitMilliseconds = 50, options = {
    isImmediate: false
}) {
    let timeoutId;
    return function (...args) {
        const context = this;
        const doLater = function () {
            timeoutId = undefined;
            if (!options.isImmediate) {
                func.apply(context, args);
            }
        };
        const shouldCallNow = options.isImmediate && timeoutId === undefined;
        if (timeoutId !== undefined) {
            clearTimeout(timeoutId);
        }
        timeoutId = setTimeout(doLater, waitMilliseconds);
        if (shouldCallNow) {
            func.apply(context, args);
        }
    };
}

function useNow(options) {
    const SYNC_MS = 1000;
    const ms = (options && options.refreshMs) || SYNC_MS;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    const fn = (options && isFunction(options.timeFn) && options.timeFn) || Date.now;
    let handler = undefined;
    let timeoutHandler = undefined;
    const now = compositionApi.ref(fn());
    const remove = () => {
        clearInterval(handler);
        clearTimeout(timeoutHandler);
    };
    /* istanbul ignore next */
    const start = isClient
        ? () => (handler = setInterval(() => (now.value = fn()), ms))
        : NO_OP;
    if (sync) {
        const offset = SYNC_MS - (now.value - Math.floor(now.value / SYNC_MS) * SYNC_MS);
        timeoutHandler = setTimeout(start, offset);
    }
    else {
        start();
    }
    compositionApi.onUnmounted(remove);
    return {
        now,
        remove
    };
}

function useDateNow(options) {
    const refreshMs = (options && options.refreshMs) || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: Date.now
    });
}

function usePerformanceNow(options) {
    const refreshMs = (options && options.refreshMs) || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: () => performance.now()
    });
}

function useFormat(format, args) {
    return compositionApi.computed(() => {
        const f = unwrap(format);
        if (!args) {
            return f;
        }
        const r = compositionApi.isRef(args) ? compositionApi.reactive(args.value) : compositionApi.reactive(args);
        const regEx = /({?{[\w\s]*}?})/g;
        return f.replace(regEx, s => {
            const k = s
                .replace("{", "")
                .replace("}", "")
                .trim();
            // ignore if {{ }}
            if (s[0] === s[1] && s[0] === "{") {
                return `${k}`;
            }
            const v = isArray(r) ? r[+k] : r[k];
            return v === undefined ? s : `${unwrap(v)}`;
        });
    });
}

function usePath(source, path, separator = ".", notFoundReturn = NO_OP) {
    return compositionApi.computed(() => {
        const s = unwrap(source);
        const p = unwrap(path);
        if (s === undefined)
            return notFoundReturn(p, s, p, s);
        if (!p) {
            return s;
        }
        const fragments = p.split(separator);
        let c = s;
        for (let i = 0; i < fragments.length; i++) {
            let fragmentPath = fragments[i];
            let index = -1;
            if (fragmentPath[fragmentPath.length - 1] === "]") {
                const m = fragmentPath.match(/\[(\d+)\]$/);
                if (m && m[1]) {
                    index = +m[1];
                    fragmentPath = fragmentPath.slice(0, -m[0].length);
                }
            }
            if (isObject(c)) {
                c = c[fragmentPath];
                // array like: when using ref with and array, it becomes arraylike object
                if (index >= 0) {
                    c = c[index];
                }
            }
            else {
                return notFoundReturn(fragments.slice(0, i).join(separator), c, p, s);
            }
            if (!c) {
                return notFoundReturn(fragments.slice(0, i).join(separator), c, p, s);
            }
        }
        return c;
    });
}

/* /Output */
function isValidation(v) {
    return typeof v.$value !== "undefined";
}
function isValidatorObject(v) {
    return isObject(v);
}
const buildValidationFunction = (r, f, handlers) => {
    const $promise = compositionApi.ref(null);
    const $pending = compositionApi.ref(false);
    const $error = compositionApi.ref();
    const $invalid = compositionApi.ref(false);
    let context = undefined;
    const onChange = (r) => {
        const p = async () => {
            try {
                $pending.value = true;
                const result = f(r, context);
                if (isPromise(result)) {
                    $invalid.value = !(await result);
                }
                else {
                    $invalid.value = !result;
                }
            }
            catch (e) {
                $invalid.value = true;
                throw e;
            }
            finally {
                $pending.value = false;
            }
        };
        $promise.value = p().catch(x => {
            $error.value = x;
            $invalid.value = true;
            return x;
        });
    };
    handlers.push((ctx) => {
        context = ctx;
        compositionApi.watch(() => {
            try {
                // keep track on the external dependencies
                f(r.value, context);
            }
            catch (e) {
                // ignore error
            }
            return r.value;
        }, onChange, { deep: true });
    });
    return {
        $promise,
        $pending,
        $invalid,
        $error
    };
};
const buildValidationValue = (r, v, handlers) => {
    const { $message, $validator, ...$rest } = isValidatorObject(v)
        ? v
        : { $validator: v, $message: "" };
    const { $pending, $promise, $invalid, $error } = buildValidationFunction(r, $validator, handlers);
    return {
        $pending,
        $error,
        $promise,
        $invalid,
        $message,
        ...$rest
    };
};
const buildValidation = (o, handlers) => {
    const r = {};
    const $value = isValidation(o) ? o.$value : undefined;
    for (const k of Object.keys(o)) {
        if (k[0] === "$") {
            if (k === "$value") {
                r[k] = $value;
                const $dirty = compositionApi.ref(false);
                const dirtyWatch = compositionApi.watch($value, () => {
                    $dirty.value = true;
                    dirtyWatch();
                }, { lazy: true, deep: true });
                r["$dirty"] = $dirty;
                continue;
            }
            else {
                r[k] = o[k];
                continue;
            }
        }
        if ($value) {
            const validation = buildValidationValue($value, o[k], handlers);
            r[k] = {
                ...validation,
                $value
            };
        }
        else {
            const validation = buildValidation(o[k], handlers);
            let $anyDirty = undefined;
            let $errors;
            let $anyInvalid;
            if (isValidation(validation)) {
                const validations = Object.keys(validation)
                    .filter(x => x[0] !== "$")
                    .map(x => validation[x]);
                $errors = compositionApi.computed(() => {
                    return validations
                        .map(x => x.$error)
                        .filter(x => Boolean(x) && Array.isArray(x) ? x.some(Boolean) : x);
                });
                // $anyDirty = computed(() => validations.some(x => !!x));
                $anyInvalid = compositionApi.computed(() => validations.some(x => !!x.$invalid));
            }
            else {
                const validations = Object.keys(validation).map(x => validation[x]);
                $errors = compositionApi.computed(() => {
                    return validations
                        .map(x => x.$errors)
                        .filter(Boolean)
                        .filter(x => x.some(Boolean));
                });
                $anyDirty = compositionApi.computed(() => validations.some(x => (x.$anyDirty && x.$anyDirty) ||
                    (isBoolean(x.$dirty) && x.$dirty)));
                $anyInvalid = compositionApi.computed(() => validations.some(x => !!x.$anyInvalid));
            }
            r[k] = {
                ...validation,
                $errors,
                $anyInvalid
            };
            if ($anyDirty) {
                r[k].$anyDirty = $anyDirty;
            }
        }
    }
    return r;
};
function useValidation(input) {
    const handlers = [];
    const validation = buildValidation({ input }, handlers);
    // convert to reactive, this will make it annoying to deconstruct, but
    // allows to use it directly on the render template without `.value`
    // https://github.com/vuejs/vue-next/pull/738
    const validationInput = compositionApi.reactive(validation.input);
    // set the context, this will allow to use this object as the second
    // argument when calling validators
    handlers.forEach(x => x(validationInput));
    return validationInput;
}

const I18n_ACCESS_SYMBOL = Symbol( ``);
function useI18n(definition) {
    if (definition) {
        return buildI18n(definition);
    }
    else
        return compositionApi.inject(I18n_ACCESS_SYMBOL);
}
function buildI18n(definition) {
    const locales = compositionApi.ref(Object.keys(definition.messages));
    const localeMessages = compositionApi.ref(definition.messages);
    const locale = compositionApi.ref(definition.locale);
    const i18n = compositionApi.ref({});
    let fallback = compositionApi.ref();
    const cache = {};
    const loadLocale = (locale, localeMessages) => {
        if (cache[locale]) {
            return cache[locale];
        }
        const l = localeMessages.value[locale];
        if (!l) {
            return compositionApi.ref({});
        }
        if (isFunction(l)) {
            return Promise.resolve(l()).then(x => (cache[locale] = wrap(x)));
        }
        return (cache[locale] = compositionApi.computed(() => localeMessages.value[locale]));
    };
    const shouldFallback = definition.fallback
        ? isBoolean(definition.notFoundFallback)
            ? definition.notFoundFallback
            : true
        : false;
    let fallbackIsPromise = false;
    if (shouldFallback) {
        const fallbackI18n = loadLocale(locale.value, localeMessages);
        if (isPromise(fallbackI18n)) {
            fallbackI18n.then(x => {
                fallback = x;
            });
            fallbackIsPromise = true;
        }
        else {
            fallback = fallbackI18n;
        }
    }
    else {
        fallback.value = {};
    }
    compositionApi.watch([locale, fallback], async ([l, fb]) => {
        if (l === definition.fallback && shouldFallback) {
            i18n.value = fb;
        }
        else {
            const localeMessage = await loadLocale(l, localeMessages);
            i18n.value = deepClone({}, fb, localeMessage.value);
        }
    }, {
        lazy: fallbackIsPromise
    });
    const $t = (path, args) => {
        if (definition.resolve) {
            return wrap(definition.resolve(i18n.value, path, args));
        }
        return useFormat(usePath(i18n, path, ".", (_, _1, p) => p), args);
    };
    const addLocale = (l, m) => {
        if (locales.value.indexOf(l) >= 0) ;
        else {
            locales.value.push(l);
        }
        delete cache[l];
        localeMessages.value[l] = m;
    };
    const removeLocale = (l) => {
        const index = locales.value.indexOf(l);
        if (index >= 0) {
            locales.value.splice(index, 1);
        }
        delete localeMessages.value[l];
        delete cache[l];
    };
    return {
        locale,
        locales,
        i18n,
        $t,
        addLocale,
        removeLocale
    };
}
function setI18n(definition) {
    const r = buildI18n(definition);
    compositionApi.provide(I18n_ACCESS_SYMBOL, r);
    return r;
}

exports.FALSE_OP = FALSE_OP;
exports.NO_OP = NO_OP;
exports.PASSIVE_EV = PASSIVE_EV;
exports.buildI18n = buildI18n;
exports.debounce = debounce;
exports.deepClone = deepClone;
exports.exponentialDelay = exponentialDelay;
exports.isArray = isArray;
exports.isBoolean = isBoolean;
exports.isClient = isClient;
exports.isDate = isDate;
exports.isElement = isElement;
exports.isFunction = isFunction;
exports.isNumber = isNumber;
exports.isObject = isObject;
exports.isPromise = isPromise;
exports.isString = isString;
exports.isSymbol = isSymbol;
exports.minMax = minMax;
exports.noDelay = noDelay;
exports.promisedTimeout = promisedTimeout;
exports.setI18n = setI18n;
exports.unwrap = unwrap;
exports.useArrayPagination = useArrayPagination;
exports.useCancellablePromise = useCancellablePromise;
exports.useDateNow = useDateNow;
exports.useDebounce = useDebounce;
exports.useFormat = useFormat;
exports.useI18n = useI18n;
exports.useNow = useNow;
exports.usePagination = usePagination;
exports.usePath = usePath;
exports.usePerformanceNow = usePerformanceNow;
exports.usePromise = usePromise;
exports.useRetry = useRetry;
exports.useValidation = useValidation;
exports.wrap = wrap;
